/*
 * ReadProvider.cpp
 *
 *  Created on: Jun 14, 2012
 *      Author: philipp_
 */

#include "ReadProvider.h"

#include <zlib.h>
#include <stdio.h>
#include <algorithm>
#include <cmath>

#include "kseq.h"
#include "Config.h"
#include "Log.h"
#include "Timing.h"
#include "CS.h"
#include "MappedRead.h"

using NGMNames::ReadStatus;

#undef module_name
#define module_name "READPROV"

KSEQ_INIT(gzFile, gzread)

gzFile fp;
kseq_t *seq;

static IRefProvider const * m_RefProvider;
static RefEntry * m_entry;
static std::map<SequenceLocation, float> iTable; // fallback

uint const estimateSize = 10000;

int * lengthTable;
int lengthTableIndex;
float * maxHitTable;
int maxHitTableIndex;
int m_CurrentReadLength;

ReadProvider::ReadProvider() {

}

inline int GetBin(uint pos) {
	static int shift = CS::calc_binshift(20);
	return pos >> shift;
	//return pos;
}

int CollectResultsFallback() {
	float max = 0;

	for (std::map<SequenceLocation, float>::iterator itr = iTable.begin(); itr != iTable.end(); itr++) {
		max = std::max(max, itr->second);
	}

	lengthTable[lengthTableIndex++] = iTable.size();
	maxHitTable[maxHitTableIndex++] = max;
	iTable.clear();
	return 0;
}

static void PrefixSearch(ulong prefix, uint pos, void* data) {

	RefEntry const * cur = m_RefProvider->GetRefEntry(prefix, m_entry); // Liefert eine liste aller Vorkommen dieses Praefixes in der Referenz

	while (cur != 0) {
		//Get kmer-weight.
//		float weight = cur->weight;
		float weight = 1.0f;

		int const n = cur->refCount;

		if (cur->reverse) {
			for (int i = 0; i < n; ++i) {
				SequenceLocation curLoc = cur->ref[i];
				curLoc.m_RefId = 1;
				curLoc.m_Location = GetBin(curLoc.m_Location - (m_CurrentReadLength - (pos + CS::prefixBasecount))); // position offset
				iTable[curLoc] += weight;
			}

		} else {
			for (int i = 0; i < n; ++i) {
				SequenceLocation curLoc = cur->ref[i];
				curLoc.m_RefId = 0;
				curLoc.m_Location = GetBin(curLoc.m_Location - pos); // position offset
				iTable[curLoc] += weight;
			}
		}

		cur = cur->nextEntry;
	}
}

uint ReadProvider::init(char const * fileName) {
	typedef void (*PrefixIterationFn)(ulong prefix, uint pos, void* data);
	PrefixIterationFn fnc = &PrefixSearch;

	bool m_EnableBS = false;
	if (Config.Exists("bs_mapping"))
		m_EnableBS = (Config.GetInt("bs_mapping", 0, 1) == 1);

	Log.Message("Initializing ReadProvider");

	uint readCount = 0;
	Timer tmr;
	tmr.ST();
	size_t maxLen = 0;
	bool estimate = !(Config.Exists("skip_estimate") && Config.GetInt("skip_estimate"));
	if (!Config.Exists("qry_max_len") || estimate) {
		if (estimate) {
			Log.Message("Estimating parameter from data.");

			lengthTable = new int[estimateSize];
			maxHitTable = new float[estimateSize];

			lengthTableIndex = 0;
			maxHitTableIndex = 0;

			m_RefProvider = NGM.GetRefProvider(0);
			m_entry = new RefEntry(0);
			m_entry->nextEntry = new RefEntry(0);
		}

		fp = gzopen(fileName, "r");
		seq = kseq_init(fp);
		int l = 0;

		size_t minLen = 9999999;
		size_t sumLen = 0;

		while ((l = kseq_read(seq)) >= 0) {
			maxLen = std::max(maxLen, seq->seq.l);
			minLen = std::min(minLen, seq->seq.l);
			sumLen += seq->seq.l;

			readCount += 1;
			if (estimate && (readCount % 100) == 0 && readCount < estimateSize) {
				m_CurrentReadLength = seq->seq.l;
				CS::PrefixIteration((char const *) seq->seq.s, (uint) seq->seq.l, fnc, (void *) this, (uint) 0, (uint) 0);
				CollectResultsFallback();
			}
		}
		maxLen = (maxLen | 1) + 1;
		int avgLen = sumLen / readCount;
		((_Config*) _config)->Override("qry_max_len", (int) maxLen);
		((_Config*) _config)->Override("qry_avg_len", (int) avgLen);
		((_Config*) _config)->Override("corridor", (int) (5 + avgLen * 0.15));
		Log.Message("Average read length: %d (min: %d, max: %d)", avgLen, minLen, maxLen);
		Log.Message("Corridor width: %d", Config.GetInt("corridor"));
		kseq_destroy(seq);
		gzclose(fp);

		if (estimate) {
			float sum = 0.0f;
			for (int i = 0; i < maxHitTableIndex; ++i) {
				sum += maxHitTable[i];
			}

			float m_CsSensitivity = 0.0f;
			if (!m_EnableBS) {
				static const int skip = (Config.Exists("kmer_skip") ? Config.GetInt("kmer_skip", 0, -1) : 0) + 1;
				float max = (avgLen - CS::prefixBasecount + 1) / skip;
				float avg = sum / maxHitTableIndex * 1.0f;
				Log.Message("Average kmer hits pro read: %f", avg);
				Log.Message("Max possible kmer hit: %f", max);

				m_CsSensitivity = (avg / (max)) * 0.90f + 0.05f;
				Log.Green("Estimated sensitivity: %f", m_CsSensitivity);

				if (Config.Exists("sensitivity")) {
					//float x = Config.GetFloat("sensitivity", 0, 100);
					//m_CsSensitivity = (100 - x) / 100.0f;
					m_CsSensitivity = Config.GetFloat("sensitivity", 0, 1);
					Log.Warning("Sensitivity threshold overwritten by user. Using %f", m_CsSensitivity);
				}
				((_Config*) _config)->Override("sensitivity", m_CsSensitivity);
			} else {
				m_CsSensitivity = 0.5f;
				if (Config.Exists("sensitivity")) {
					m_CsSensitivity = Config.GetFloat("sensitivity", 0, 1);
				}
			}

			//std::sort(lengthTable, lengthTable + lengthTableIndex);
			//int size = lengthTable[(int) (0.5f * lengthTableIndex) - 1] * 32;
			//int bits = (int) std::min(20.0, ceil(log2(size)));
			//Log.Green("Optimale hash table size: %d (%d)", (int)pow(2.0, (double)bits), lengthTable[lengthTableIndex - 1]);
			//if(Config.Exists("cs_tablen")) {
			//	bits = Config.GetInt("cs_tablen", 0, 32);
			//	Log.Warning("SearchTableLength overwritten by user. Using %d bits", (int)pow(2.0, (double)bits));
			//}
			//((_Config*) _config)->Override("cs_tablen", bits);

			delete[] lengthTable;
			delete[] maxHitTable;
		}

	} else {
		Log.Warning("qry_max_len was found in config file. Please make sure that this number is correct: %d", Config.GetInt("qry_max_len"));
	}

	if (!Config.Exists("sensitivity")) {
		((_Config*) _config)->Override("sensitivity", 0.5f);
		Log.Warning("Sensitivity parameter neither set nor estimated. Falling back to default.");
	}

	fp = gzopen(fileName, "r");
	seq = kseq_init(fp);

	Log.Message("Initializing took %.3fs", tmr.ET());

	return readCount;
}

ReadProvider::~ReadProvider() {
	kseq_destroy(seq);
	gzclose(fp);
}

MappedRead * ReadProvider::nextRead(int const id) {
	static int const qryMaxLen = Config.GetInt("qry_max_len");
	MappedRead * read = 0;
	int l = kseq_read(seq);
//	Log.Message("Name (%d): %s", seq->name.l, seq->name.s);
//	Log.Message("Seq  (%d): %s", seq->seq.l, seq->seq.s);
//	Log.Message("Qual (%d): %s", seq->qual.l, seq->qual.s);
	if (l >= 0) {
		if (seq->seq.l == seq->qual.l || seq->qual.l == 0) {
			read = new MappedRead(id);

			//Name
			static size_t const MAX_READNAME_LENGTH = 100;
			read->name = new char[MAX_READNAME_LENGTH];
			read->nameLength = std::min(MAX_READNAME_LENGTH, seq->name.l);
			memcpy(read->name, seq->name.s, read->nameLength);

			//Sequence
			read->Seq = new char[qryMaxLen];
			memset(read->Seq, '\0', qryMaxLen);
			read->length = seq->seq.l;
			int nCount = 0;
			for (int i = 0; i < read->length; ++i) {
				char c = toupper(seq->seq.s[i]);
				if (c == 'A' || c == 'T' || c == 'C' || c == 'G') {
					read->Seq[i] = c;
				} else {
					read->Seq[i] = 'N';
					nCount += 1;
				}

			}
//			if (nCount > qryMaxLen * 0.5f) {
//				Log.Warning("Discarding read %s (too many Ns)", read->name);
//				delete read;
//				return 0;
//			}
//			for (int i = read->length; i < qryMaxLen; ++i) {
//				read->Seq[i] = '\0';
//			}

			//memcpy(read->Seq, seq->seq.s, read->length);

			//Quality
			read->qlty = 0;
			if (seq->qual.l > 0) {
				read->qlty = new char[qryMaxLen];
				memcpy(read->qlty, seq->qual.s, read->length);
			}

//			Log.Message("%s", read->name);
//			Log.Message("%s", read->Seq);
//			if (read->qlty != 0)
//				Log.Message("%s", read->qlty);
		} else {
			Log.Warning("Discarding read %s. Length of read not equal length of quality values.", seq->name.s);
		}
	}
	return read;
}

//bool _SequenceProvider::IsValid(int n) const
//{
//	CheckQryNr(n);
//
//	if (NGM.Paired())
//	{
//		bool offstrand = (n & 1);
//		n >>= 1;
//
//		if (offstrand)
//			return prdIdx[n].Flags & 0x1;
//	}
//
//	return qryIdx[n].Flags & 0x1;
//}

MappedRead * ReadProvider::GenerateSingleRead(int const readid) {
	MappedRead * read = 0;

//	if (SequenceProvider.IsValid(readid))
	read = nextRead(readid); //

	return read;
}

// Sequential (important for pairs!) read generation
MappedRead * ReadProvider::GenerateRead(int const readid) {
	static MappedRead * nextRead = 0;

	if (NGM.Paired()) {
		MappedRead * read = 0;
		if (!(readid & 0x1)) // Requested first read of a pair (even-numbered read id)
		{
			// Generate whole pair
			read = GenerateSingleRead(readid);
			nextRead = GenerateSingleRead(readid + 1);

			if (read != 0) {
				read->Paired = nextRead;
				if (nextRead == 0)
					read->SetFlag(NGMNames::NoSrcPair);
			}
			if (nextRead != 0) {
				nextRead->Paired = read;
				if (read == 0)
					nextRead->SetFlag(NGMNames::NoSrcPair);
			}
		} else {
			read = nextRead;
			nextRead = 0;

			if ((read != 0) && (read->ReadId != readid)) {
				Log.Error("Non-sequential read generation (req %i, got %i)", readid, read->ReadId);
				Fatal();
			}

		}

		return read;
	} else {
		return GenerateSingleRead(readid);
	}
}

void ReadProvider::DisposeRead(MappedRead * read) {
	if (NGM.Paired() && read->Paired != 0) // Program runs in paired mode and pair exists
			{
		if (read->Paired->HasFlag(NGMNames::DeletionPending)) // Paired read was marked for deletion
				{
			// delete paired and read
			Log.Verbose("Deleting read %d (%s)", read->ReadId, read->name);
			Log.Verbose("Deleting read %d (%s)", read->Paired->ReadId, read->Paired->name);
			delete read->Paired;
			delete read;
		} else {
			// Paired read is still in use...mark read for deletion
			read->SetFlag(NGMNames::DeletionPending);
		}
	} else {
		// Single mode or no existing pair
		Log.Verbose("Deleting single read %d (%s)", read->ReadId, read->name);
		delete read;
	}
}
